package org.t2framework.lucy;

import java.util.List;

import org.t2framework.commons.meta.BeanDesc;
import org.t2framework.commons.meta.DescType;
import org.t2framework.commons.meta.MethodDesc;
import org.t2framework.commons.meta.PropertyDesc;
import org.t2framework.commons.util.Assertion;
import org.t2framework.lucy.exception.BehaviorNotFoundException;
import org.t2framework.lucy.spi.Behavior;
import org.t2framework.lucy.spi.ComponentBuilder;
import org.t2framework.lucy.spi.LucyConfiguration;

/**
 * ComponentBuilderFactory creates ComponentBuilder.Unlike it is,
 * ComponentBuilderFactory should be extensible for some reason, but not now.
 * 
 * @author shot
 */
public class ComponentBuilderFactory {

	public static ComponentBuilder createBuilder(BeanDesc<?> beanDesc, Lucy lucy) {
		return new DefaultComponentBuilderImpl(beanDesc, lucy);
	}

	public static ComponentBuilder createBuilder(BeanDesc<?> beanDesc,
			Lucy lucy, LucyConfiguration conf) {
		return createBuilder(beanDesc, lucy).loadConfiguration(conf);
	}

	public static class DefaultComponentBuilderImpl implements ComponentBuilder {

		protected BeanDesc<?> beanDesc;

		protected Lucy lucy;

		protected LucyConfiguration conf;

		public DefaultComponentBuilderImpl(BeanDesc<?> beanDesc, Lucy lucy) {
			this.beanDesc = beanDesc;
			this.lucy = lucy;
		}

		@Override
		public <T> T build(T t, ConfigBindResult result) {
			Assertion.notNull(result);
			for (Lifecycle current : Lifecycle.getComponentLifecycle()) {
				t = processClassDesc(t, result, current);
				processMethodAndProperties(t, result, current);
			}
			return t;
		}

		protected <T> T processClassDesc(T t, ConfigBindResult result,
				Lifecycle current) {
			return processBehavior(t, result, current, beanDesc.getClassDesc());
		}

		protected <T> T processBehavior(T t, ConfigBindResult result,
				Lifecycle current, DescType desc) {
			List<Behavior> behaviorList = result.getBehaviorList(desc);
			if (behaviorList == null) {
				throw new BehaviorNotFoundException(t, current, desc);
			}
			for (Behavior behavior : behaviorList) {
				if (behavior.isApplied(current)) {
					t = behavior.execute(t, lucy);
				}
			}
			return t;
		}

		protected <T> T processMethodAndProperties(T t,
				ConfigBindResult result, Lifecycle current) {
			for (MethodDesc md : beanDesc.getAllMethodDesc()) {
				if (result.hasMethodDesc(md) == false) {
					continue;
				}
				t = processBehavior(t, result, current, md);
			}
			for (PropertyDesc<?> pd : beanDesc.getAllPropertyDesc()) {
				if (result.hasPropertyDesc(pd) == false) {
					continue;
				}
				t = processBehavior(t, result, current, pd);
			}
			return t;
		}

		@Override
		public void setLucy(Lucy lucy) {
			this.lucy = lucy;
		}

		@Override
		public void dispose() {
			beanDesc = null;
			lucy = null;
		}

		@Override
		public ComponentBuilder loadConfiguration(LucyConfiguration conf) {
			this.conf = Assertion.notNull(conf);
			return this;
		}

	}
}
