package org.t2framework.lucy;

import org.t2framework.commons.exception.NoSuchComponentException;
import org.t2framework.commons.meta.BeanDesc;
import org.t2framework.commons.meta.PackageDescHolder;
import org.t2framework.lucy.exception.CyclicReferenceRuntimeException;
import org.t2framework.lucy.exception.TooManyRegistrationException;

/**
 * {@.en A simple IoC container without difficult things.}
 * 
 * <br />
 * 
 * {@.ja LucyはシンプルなIoCコンテナです.}
 * 
 * @author shot
 */
public interface Lucy extends PackageDescHolder, ComponentLoader {

	/**
	 * {@.en Initialize {@link Lucy}.}
	 * 
	 * <br />
	 * 
	 * {@.ja Lucyを初期化します.}
	 */
	void init();

	/**
	 * {@.en Destroy this Lucy instance.}
	 * 
	 * <br />
	 * 
	 * {@.ja Lucyを破棄します.Lucyが保持する開放されるべきリソースも同時に開放されます.}
	 */
	void destroy();

	/**
	 * {@.en Register the class. The class must not be null.}
	 * 
	 * <br />
	 * 
	 * {@.ja Lucyに渡されたClassを登録します.ClassはNullであってはいけません.}
	 * 
	 * @param <T>
	 * @param componentClass
	 *            (must not be null)
	 * @return {@link Lucy}
	 */
	<T> Lucy register(Class<? extends T> componentClass);

	/**
	 * {@.en Register the instance.The instance must not be null.}
	 * 
	 * <br />
	 * 
	 * {@.ja Lucyに渡されたインスタンスを登録します.インスタンスはnullであってはいけません.}
	 * 
	 * @param <T>
	 * @param t
	 * @return {@link Lucy}
	 */
	<T> Lucy register(T t);

	/**
	 * {@.en Register class with an identifier.}
	 * 
	 * <br />
	 * 
	 * {@.ja Classを識別子と共に登録します.Class、識別子共にnullであってはいけません.}
	 * 
	 * @param <T>
	 * @param componentClass
	 * @param identifier
	 * @return
	 */
	<T> Lucy register(Class<? extends T> componentClass, Object identifier);

	/**
	 * {@.en Register an instance with an identifier.Both instance and
	 * identifier must not be null.}
	 * 
	 * <br />
	 * 
	 * {@.ja インスタンスを識別子と共に登録します.インスタンス、識別子共にnullであってはいけません.}
	 * 
	 * @param <T>
	 * @param t
	 *            (must not be null)
	 * @param identifier
	 *            (must not be null)
	 * @return {@link Lucy}
	 */
	<T> Lucy register(T t, Object identifier);

	/**
	 * {@.en Register {@link BeanDesc}, a descriptor class of a bean.}
	 * 
	 * <br />
	 * 
	 * {@.ja Beanの記述情報クラスである、BeanDescを登録します.BeanDescはnullであってはいけません.}
	 * 
	 * @param <T>
	 * @param beanDesc
	 *            (must not be null)
	 * @return {@link Lucy}
	 */
	<T> Lucy register(BeanDesc<? extends T> beanDesc);

	/**
	 * 
	 * {@.en Register {@link BeanDesc} with an identifier.}
	 * 
	 * <br />
	 * 
	 * {@.ja {@code BeanDesc}を識別子つきで登録します．}
	 * 
	 * @param <T>
	 * @param beanDesc
	 *            (must not be null)
	 * @param identifier
	 *            (must not be null)
	 * @return
	 */
	<T> Lucy register(BeanDesc<? extends T> beanDesc, Object identifier);

	/**
	 * 
	 * {@.en Get component by class.If {@link Lucy} has more than one
	 * component, it will throw {@link TooManyRegistrationException} by
	 * default.}
	 * 
	 * <br />
	 * 
	 * {@.ja {@link Class}をキーにコンポーネントを取得します.}
	 * 
	 * @param <T>
	 * @param key
	 * @return
	 * @throws TooManyRegistrationException
	 * @throws CyclicReferenceRuntimeException
	 * @throws NoSuchComponentException
	 */
	<T> T get(Class<? super T> key) throws TooManyRegistrationException,
			CyclicReferenceRuntimeException, NoSuchComponentException;

	/**
	 * 
	 * {@.en Get component by identifier. }
	 * 
	 * <br />
	 * 
	 * {@.ja {@link Object}をキーにコンポーネントを取得します.}
	 * 
	 * @param <T>
	 * @param identifier
	 * @return
	 */
	<T> T get(Object identifier);

	/**
	 * 
	 * {@.en Get all components by class.This method does not throw {
	 * @link TooManyListenersException} in design and will get multiple
	 * components. }
	 * 
	 * <br />
	 * 
	 * {@.ja {@link Class}をキーに取得できるコンポーネントの配列を返します.キーで登録されているコンポーネントが
	 * 無い場合、空の配列を返します.}
	 * 
	 * @param <T>
	 * @param key
	 * @return
	 * @throws CyclicReferenceRuntimeException
	 */
	<T> T[] getAll(Class<? super T> key) throws CyclicReferenceRuntimeException;

	/**
	 * 
	 * {@.en Get BeanDesc by identifier. }
	 * 
	 * <br />
	 * 
	 * {@.ja オブジェクトの識別子をキーに{@link BeanDesc}を取得します.
	 * キーで登録されているコンポーネントが無い場合、デフォルトではnullを返します. コンポーネントが無い場合の挙動は
	 * {@link BeanDescStrategy}で変更することが出来ます.}
	 * 
	 * @param <T>
	 * @param identifier
	 * @return
	 * @throws CyclicReferenceRuntimeException
	 */
	<T> BeanDesc<T> getBeanDesc(Object identifier)
			throws CyclicReferenceRuntimeException;

	/**
	 * 
	 * {@.en Get array of BeanDesc by class.This method does not throw {
	 * @link TooManyRegistrationException} in design and will get multiple
	 * components. }
	 * 
	 * <br />
	 * 
	 * {@.ja {@link Class}をキーに{@link BeanDesc}の配列を返します.このメソッドは
	 * {@link TooManyRegistrationException}は発生しません. {@link Class}
	 * がキーで登録されているコンポーネントが無い場合、空の配列を返します.}
	 * 
	 * @param <T>
	 * @param key
	 * @return
	 * @throws CyclicReferenceRuntimeException
	 */
	<T> BeanDesc<T>[] getBeanDescs(Class<? super T> key)
			throws CyclicReferenceRuntimeException;

	/**
	 * 
	 * {@.en Get {@link BeanDesc} by class.}
	 * 
	 * <br />
	 * 
	 * {@.ja {@link Class}をキーにして{@link BeanDesc}を取得します.}
	 * 
	 * @param <T>
	 * @param key
	 * @return
	 * @throws TooManyRegistrationException
	 * @throws CyclicReferenceRuntimeException
	 */
	<T> BeanDesc<T> getBeanDesc(Class<? super T> key)
			throws TooManyRegistrationException,
			CyclicReferenceRuntimeException;

	/**
	 * 
	 * {@.en True if a component exists by the class. }
	 * 
	 * <br />
	 * 
	 * {@.ja Classにマッチするコンポーネントが存在するかチェックします.存在する場合、trueを返します.}
	 * 
	 * @param <T>
	 * @param key
	 * @return
	 */
	<T> boolean hasComponent(Class<T> key);

	/**
	 * 
	 * {@.en True if a component exists by identifier. }
	 * 
	 * <br />
	 * 
	 * {@.ja 識別子にマッチするコンポーネントが存在するかチェックします.存在する場合、trueを返します.}
	 * 
	 * @param identifier
	 * @return
	 */
	boolean hasComponent(Object identifier);

	/**
	 * 
	 * {@.en Inject dependency. }
	 * 
	 * <br />
	 * 
	 * {@.ja インスタンスTに対して、依存解決をします.}
	 * 
	 * @param <T>
	 * @param t
	 * @return
	 */
	<T> T injectDependency(T t);

}
