package org.t2framework.lucy.aop.behavior;

import java.util.List;

import org.t2framework.commons.annotation.ConfigurationTarget;
import org.t2framework.commons.aop.AopEngine;
import org.t2framework.commons.aop.AspectAssembler;
import org.t2framework.commons.aop.spi.Aspect;
import org.t2framework.commons.meta.BeanDesc;
import org.t2framework.commons.meta.ClassDesc;
import org.t2framework.commons.meta.Config;
import org.t2framework.commons.meta.InstanceFactory;
import org.t2framework.commons.module.Modules;
import org.t2framework.lucy.Lifecycle;
import org.t2framework.lucy.Lucy;
import org.t2framework.lucy.aop.config.AspectConfig;
import org.t2framework.lucy.spi.AbstractBehavior;

/**
 * AspectBehavior is a behavior class for enhancing component. Enhancing
 * component executes only one time, and it will be cached as {@link BeanDesc}.
 * 
 * The target configurations are
 * {@link org.t2framework.commons.annotation.core.Aspect} for annotation one,
 * and {@link AspectConfig} for xml one.
 * 
 * @author shot
 */
@ConfigurationTarget( { org.t2framework.commons.annotation.core.Aspect.class,
		AspectConfig.class })
public abstract class AbstractAspectBehavior extends AbstractBehavior {

	protected AopEngine aopEngine;

	protected Object lock = new Object();

	@SuppressWarnings("unchecked")
	public AbstractAspectBehavior(BeanDesc beanDesc, ClassDesc classDesc,
			Config cd) {
		super(beanDesc, classDesc, cd);
	}

	@Override
	public <T> void init() {
		super.init();
		aopEngine = Modules.getService(AopEngine.class);
	}

	@Override
	public <T> T execute(T t, Lucy lucy) {
		// @Aspect handles just once, not twice or more.
		BeanDesc<Object> beanDesc = getBeanDesc();
		synchronized (beanDesc) {
			if (beanDesc.getEnhancedComponentClass() != null) {
				return t;
			}
		}
		Class<? extends Object> originalClass = beanDesc.getComponentClass();
		AspectAssembler<Lucy> aspectAssembler = createAspectAssembler();
		List<Aspect> aspects = aspectAssembler.assemble(beanDesc, lucy);

		Class<? extends Object> enhancedClass = aopEngine.getEnhancedClass(
				originalClass, aspects);
		synchronized (beanDesc) {
			beanDesc.setEnhancedComponentClass(enhancedClass);
		}

		InstanceFactory<Object> instanceFactory = createInstanceFactory(aspects);
		beanDesc.setInstanceFactory(instanceFactory);

		return t;
	}

	protected abstract AspectAssembler<Lucy> createAspectAssembler();

	protected abstract InstanceFactory<Object> createInstanceFactory(
			List<Aspect> aspects);

	@Override
	public <T> void destroy() {
		synchronized (lock) {
			aopEngine.shutdown();
			aopEngine = null;
		}
		super.destroy();
	}

	@Override
	public Lifecycle getLifecycle() {
		return Lifecycle.CLASS_LOADING;
	}

}
