package org.t2framework.lucy.behavior;

import org.t2framework.commons.annotation.ConfigurationTarget;
import org.t2framework.commons.meta.BeanDesc;
import org.t2framework.commons.meta.Config;
import org.t2framework.commons.meta.MethodDesc;
import org.t2framework.commons.meta.PropertyDesc;
import org.t2framework.commons.util.Assertion;
import org.t2framework.commons.util.Logger;
import org.t2framework.lucy.Lifecycle;
import org.t2framework.lucy.Lucy;
import org.t2framework.lucy.annotation.core.Inject;
import org.t2framework.lucy.config.meta.InjectConfig;
import org.t2framework.lucy.inject.Injector;
import org.t2framework.lucy.inject.MethodInjector;
import org.t2framework.lucy.inject.PropertyInjector;
import org.t2framework.lucy.spi.AbstractBehavior;
import org.t2framework.lucy.spi.Behavior;

/**
 * 
 * {@.en Inject behavior is the behavior class to inject component.}
 * 
 * <br />
 * 
 * {@.ja {@link InjectBehavior}はコンポーネントへのインジェクトを実現する{@link Behavior}です.}
 * 
 * @author shot
 * 
 */
@ConfigurationTarget( { Inject.class, InjectConfig.class })
public class InjectBehavior extends AbstractBehavior {

	protected static Logger logger = Logger.getLogger(InjectBehavior.class);

	protected final Injector injector;

	protected Object lock = new Object();

	/**
	 * {@.en Create {@link InjectBehavior} with {@link MethodDesc}. Method
	 * injection will be executed.}
	 * 
	 * <br />
	 * 
	 * {@.ja {@link InjectBehavior}を{@link MethodDesc}と共にインスタンス化します．
	 * {@link #execute(Object, Lucy)}でメソッドインジェクションが実行されます．}
	 * 
	 * @param beanDesc
	 * @param md
	 * @param cd
	 */
	@SuppressWarnings("unchecked")
	public InjectBehavior(BeanDesc beanDesc, MethodDesc md, Config cd) {
		super(beanDesc, md, cd);
		injector = new MethodInjector(md);
	}

	/**
	 * {@.en Create {@link InjectBehavior} with {@link PropertyDesc}. Property
	 * injection will be executed.}
	 * 
	 * <br />
	 * 
	 * {@.ja {@link InjectBehavior}を{@link PropertyDesc}と共にインスタンス化します．
	 * {@link #execute(Object, Lucy)}でプロパティインジェクションが実行されます．}
	 * 
	 * @param beanDesc
	 * @param md
	 * @param cd
	 */
	@SuppressWarnings("unchecked")
	public InjectBehavior(BeanDesc beanDesc, PropertyDesc pd, Config cd) {
		super(beanDesc, pd, cd);
		injector = new PropertyInjector(pd);
	}

	/**
	 * <pre>
	 * Inject components to given parameter as t using Lucy.
	 * Execute method is called every time when @Inject is targeted in a component.
	 * For example, a component has two @Inject methods, so there should be two time execute called.
	 * 
	 * [Features of return value]
	 *  The return value must be injected components, which is targeted by @Inject 
	 *  or inject tag by xml file normally. If fail, no components injected.
	 *  
	 * [In case of fail]
	 *  In case of injecting fail, no components injected by this behavior.
	 *  
	 * [Etc]
	 *  It might work with static method injection, but not for sure.
	 * 
	 * </pre>
	 * 
	 * @param t
	 *            : target instance to be injected
	 * @param lucy
	 *            : IoC container
	 * @return t : object which may be injected some components
	 * @throws NullPointerException
	 *             in case of {@code Lucy} instance is {@code null}.
	 * @see org.t2framework.lucy.Lucy
	 */
	@Override
	public <T> T execute(T t, Lucy lucy) {
		Assertion.notNull(lucy);
		synchronized (lock) {
			injector.inject(t, lucy);
		}
		return t;
	}

	@Override
	public Lifecycle getLifecycle() {
		return Lifecycle.COMPONENT_CREATED;
	}

}
