package org.t2framework.lucy.config.stax;

import java.util.List;

import org.t2framework.commons.meta.BeanDesc;
import org.t2framework.commons.meta.Config;
import org.t2framework.commons.meta.ConfigContainer;
import org.t2framework.commons.meta.MethodDesc;
import org.t2framework.commons.util.CollectionsUtil;
import org.t2framework.commons.util.StringUtil;
import org.t2framework.commons.util.Reflections.ClassUtil;
import org.t2framework.lucy.config.meta.impl.LifeCycleMethodConfig;
import org.t2framework.lucy.exception.ArgumentMismatchException;
import org.t2framework.lucy.exception.IllegalELResultException;
import org.t2framework.lucy.exception.MethodNotFoundException;

/**
 * 
 * {@.en Base life cycle tag handler, such as {@link InitTagHandler} or
 * {@link DestroyTagHandler}.}
 * 
 * <br />
 * 
 * {@.ja }
 * 
 * @author shot
 * @author c9katayama
 * 
 */
public abstract class LifeCycleTagHandler extends AbstractExpressionHandler
		implements XmlEventHandler {

	protected abstract LifeCycleMethodConfig createConfig(
			ConfigContainer configContainer);

	@Override
	public void start(XmlEventContext context, Attributes attributes) {
		String name = attributes.getValue("method");
		if (StringUtil.isEmpty(name) == false) {
			context.push(name);
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public void end(XmlEventContext context, String body) {
		final String name = (String) context.pop();
		BeanDesc<?> beanDesc = (BeanDesc<?>) context.peek();
		String s = body.trim();
		if (StringUtil.isEmpty(s)) {
			setupNoArgMethodDesc(context, beanDesc, name);
		} else {
			final List<MethodDesc> list = beanDesc.getMethodDescs(name);
			Object o = evaluateExpression(context, s);
			if (o == null) {
				throw new IllegalELResultException(s, o);
			}
			if (o.getClass().isArray()) {
				final Object[] args = (Object[]) o;
				setupMethodDesc(beanDesc, args, list);
			} else if (o instanceof List) {
				final Object[] args = ((List) o).toArray();
				setupMethodDesc(beanDesc, args, list);
			} else {
				setupSingleArgMethodDesc(beanDesc, o, list);
			}
		}
	}

	protected void setupSingleArgMethodDesc(final BeanDesc<?> beanDesc,
			Object o, List<MethodDesc> list) {
		final List<MethodDesc> candidates = filterByParameterSize(1, list);
		for (MethodDesc md : candidates) {
			final Class<?>[] types = md.getParameterTypes();
			if (types.length != 1) {
				continue;
			}
			Class<?> type = ClassUtil.toWrapperClass(types[0]);
			if (!type.isAssignableFrom(o.getClass())) {
				continue;
			}
			final LifeCycleMethodConfig cd = createConfig(md);
			cd.setArgs(new Object[] { o });
			beanDesc.addConfig(cd);
		}
	}

	protected void setupMethodDesc(final BeanDesc<?> beanDesc,
			final Object[] args, final List<MethodDesc> list) {
		final int length = args.length;
		final List<MethodDesc> candidates = filterByParameterSize(length, list);
		if (candidates.size() == 1) {
			MethodDesc md = candidates.get(0);
			final LifeCycleMethodConfig cd = createConfig(md);
			cd.setArgs(args);
			beanDesc.addConfig(cd);
			return;
		}
		outer: for (MethodDesc md : candidates) {
			final Class<?>[] types = md.getParameterTypes();
			for (int i = 0; i < length; i++) {
				Object o = args[i];
				Class<?> type = ClassUtil.toWrapperClass(types[i]);
				if (!type.isAssignableFrom(o.getClass())) {
					continue outer;
				}
			}
			final LifeCycleMethodConfig cd = createConfig(md);
			cd.setArgs(args);
			beanDesc.addConfig(cd);
		}
	}

	protected List<MethodDesc> filterByParameterSize(final int length,
			final List<MethodDesc> list) {
		List<MethodDesc> candidates = CollectionsUtil.newArrayList();
		for (MethodDesc md : list) {
			if (md.getParameterTypes().length == length) {
				candidates.add(md);
			}
		}
		return candidates;
	}

	protected void setupNoArgMethodDesc(XmlEventContext context,
			BeanDesc<?> beanDesc, String name) {
		MethodDesc md = find(context, beanDesc, name);
		Config config = createConfig(md);
		beanDesc.addConfig(config);
	}

	protected MethodDesc find(XmlEventContext context, BeanDesc<?> bd,
			String name) {
		List<MethodDesc> list = bd.getMethodDescs(name);
		if (list == null || list.size() < 1) {
			throw new MethodNotFoundException(bd.getComponentClass(), name,
					context.getLineNumber(), context.getColumnNumber());
		}
		for (MethodDesc md : list) {
			Class<?>[] ary = md.getMethod().getParameterTypes();
			if (ary == null || ary.length == 0) {
				return md;
			}
		}
		throw new ArgumentMismatchException(bd.getComponentClass(), name,
				context.getLineNumber(), context.getColumnNumber());
	}
}
