package org.t2framework.lucy.inject;

import org.t2framework.commons.exception.IllegalAccessRuntimeException;
import org.t2framework.commons.exception.InvocationTargetRuntimeException;
import org.t2framework.commons.meta.MethodDesc;
import org.t2framework.commons.util.Logger;
import org.t2framework.commons.util.PrintableUtil;
import org.t2framework.lucy.Lucy;
import org.t2framework.lucy.exception.CyclicReferenceRuntimeException;

/**
 * 
 * {@.en {@link MethodInjector} is a concrete {@link Injector} class for method
 * injection.
 * 
 * <pre>
 * MethodInjection is enables to inject these type of injection:
 *  - simply single argument such as void inject(Emp emp);
 *  - single array arguments such as void inject(Emp[] emp);
 *  - multiple arguments not including array ex. void inject(Emp emp, Dept dept);
 * </pre>
 * 
 * }
 * 
 * <br />
 * 
 * {@.ja メソッドインジェクションを実現します．
 * 
 * 
 * メソッドインジェクションでは下記のようなインジェクトが可能です:
 * 
 * <pre>
 *  - 単一引数のインジェクト（例：void inject(Emp emp);）
 *  - 配列を使った同一の型を持つ複数コンポーネントのインジェクト（例：void inject(Emp[] emp);）
 *  - 配列を含まない複数の引数の同時インジェクト（例：void inject(Emp emp, Dept dept);）
 * </pre>
 * }
 * 
 * @author shot
 * 
 */
public class MethodInjector extends AbstractInjector<MethodDesc> {

	protected static Logger logger = Logger.getLogger(MethodInjector.class);

	protected final String targetClassName;

	protected final String targetMethodName;

	public MethodInjector(MethodDesc methodDesc) {
		super(methodDesc);
		this.targetClassName = methodDesc.getDeclaringClass().getName();
		this.targetMethodName = methodDesc.getMethodName();
	}

	/**
	 * Inject to the instance using with {@link MethodDesc}.
	 */
	@Override
	public <T> T inject(T t, Lucy lucy) throws CyclicReferenceRuntimeException,
			InvocationTargetRuntimeException, IllegalAccessRuntimeException {
		final MethodDesc methodDesc = getConfigContainer();
		final InjectInfo info = getInjectInfo();
		final Class<?>[] types = getParameterTypes(methodDesc, info);
		final Object[] args = prepareArgs(types, methodDesc, lucy, info);
		injectMethodDesc(methodDesc, t, args, types);
		return t;
	}

	/**
	 * Prepare method invocation arguments by expression if {@link MethodDesc}
	 * has, or by types and {@link Lucy} container.
	 * 
	 * @param types
	 * @param methodDesc
	 * @param lucy
	 * @return
	 */
	protected Object[] prepareArgs(Class<?>[] types, MethodDesc methodDesc,
			Lucy lucy, InjectInfo info) {
		Object[] args = null;
		if (methodDesc.hasExpression()) {
			args = evaluateExpression(methodDesc, lucy);
		} else {
			args = super.prepareArgsByTypes(types, lucy, info);
		}
		return args;
	}

	/**
	 * Method injection execution.
	 * 
	 * @param <T>
	 * @param methodDesc
	 * @param t
	 * @param args
	 * @param types
	 */
	protected <T> void injectMethodDesc(MethodDesc methodDesc, T t,
			Object[] args, Class<?>[] types) {
		if (args == null) {
			logForNullInject(types);
			return;
		}
		if (1 < args.length || (args.length == 1 && args[0] != null)) {
			methodDesc.invoke(t, args);
		} else {
			logForNullInject(types);
		}
	}

	protected void logForNullInject(Class<?>[] types) {
		logger.log("DLucyCore0001", targetClassName, targetMethodName,
				PrintableUtil.toPrintableString(types));
	}

	protected Class<?>[] getParameterTypes(MethodDesc methodDesc,
			InjectInfo info) {
		return info.hasParameterTypes() ? info.getParameterTypes() : methodDesc
				.getParameterTypes();
	}

}
