package org.t2framework.lucy.inject;

import org.t2framework.commons.exception.IllegalPropertyRuntimeException;
import org.t2framework.commons.meta.PropertyDesc;
import org.t2framework.commons.util.Logger;
import org.t2framework.lucy.Lucy;
import org.t2framework.lucy.exception.CyclicReferenceRuntimeException;

/**
 * 
 * {@.en {@link PropertyInjector} is a concrete {@link Injector} class for
 * JavaBean's property injection. }
 * 
 * <br />
 * 
 * {@.ja JavaBeans仕様に基づくプロパティを使ったプロパティインジェクションを実現します．}
 * 
 * @author shot
 * 
 */
@SuppressWarnings("unchecked")
public class PropertyInjector extends AbstractInjector<PropertyDesc<?>> {

	protected static Logger logger = Logger.getLogger(PropertyInjector.class);

	protected final String targetClassName;

	protected final String targetWriteMethodName;

	public PropertyInjector(PropertyDesc<?> propertyDesc) {
		super(propertyDesc);
		this.targetClassName = propertyDesc.getTargetClass().getName();
		if (!propertyDesc.isWritable()) {
			throw new IllegalStateException();
		}
		this.targetWriteMethodName = propertyDesc.getWriteMethod().getName();
	}

	@Override
	public <T> T inject(T t, Lucy lucy) throws CyclicReferenceRuntimeException,
			IllegalPropertyRuntimeException {
		final PropertyDesc propertyDesc = getConfigContainer();
		final InjectInfo info = getInjectInfo();
		final Class type = propertyDesc.getPropertyType();
		final Object o = prepareArg(type, propertyDesc, lucy, info);
		injectPropertyDesc(propertyDesc, t, o);
		return t;
	}

	protected <T> void injectPropertyDesc(PropertyDesc propertyDesc, T t,
			Object arg) {
		if (arg != null) {
			propertyDesc.setValue(t, arg);
		} else {
			Class type = propertyDesc.getPropertyType();
			logForNullInject(type, type.isArray());
		}
	}

	protected Object prepareArg(final Class<?> type,
			final PropertyDesc<?> propertyDesc, final Lucy lucy,
			final InjectInfo info) {
		Object o = null;
		if (type.isArray() == false) {
			o = prepareComponent(propertyDesc, type, lucy, info);
		} else {
			o = prepareComponents(propertyDesc, type, lucy, info);
		}
		return o;
	}

	protected void logForNullInject(Class<?> type, boolean isArray) {
		StringBuilder builder = new StringBuilder();
		builder.append(type.getName());
		builder.append((isArray ? "[]" : ""));
		logger.log("DLucyCore0002", targetClassName, targetWriteMethodName,
				new String(builder));
	}

	protected Object prepareComponent(PropertyDesc<?> propertyDesc,
			Class<?> type, Lucy lucy, InjectInfo info) {
		Object o = null;
		if (propertyDesc.hasExpression()) {
			Object[] evalResult = evaluateExpression(propertyDesc, lucy);
			if (evalResult != null && 0 < evalResult.length) {
				o = evalResult[0];
			} else {
				o = null;
			}
		} else {
			o = (info.hasNames()) ? lucy.get(info.getNames()[0]) : lucy
					.get(type);
		}
		return o;
	}

	protected Object prepareComponents(PropertyDesc<?> propertyDesc,
			Class<?> type, Lucy lucy, InjectInfo info) {
		if (type.isArray() == false) {
			return null;
		}
		Object o = null;
		final Class<?>[] injectTypes = InjectorUtil.getTypes(propertyDesc);
		final Class componentType = type.getComponentType();
		if (injectTypes != null) {
			o = super.prepareArgsByTypes(injectTypes, lucy, info);
		} else {
			if (propertyDesc.hasExpression()) {
				Object[] evalResult = evaluateExpression(propertyDesc, lucy);
				if (evalResult != null && 0 < evalResult.length) {
					o = convertTypes(evalResult[0], componentType);
				} else {
					o = null;
				}
			} else {
				o = lucy.getAll(componentType);
			}
		}
		return o;
	}

}
